import time
from abc import ABCMeta, abstractmethod
from logging import Logger
from threading import Thread
from typing import Optional

from asyncio_inspector.stats import BaseStatsTracker


class BaseReporter(metaclass=ABCMeta):
    stats_tracker: Optional[BaseStatsTracker] = None

    @abstractmethod
    def start(self):
        ...

    @abstractmethod
    def stop(self):
        ...


class LoggerReporter(Thread, BaseReporter):
    """Reporter that returns"""

    logger: Logger
    running: bool
    sleep_period: int
    show_top_n: int

    def __init__(self, logger: Logger) -> None:
        super().__init__()
        self.logger = logger
        self.running = True
        self.sleep_period = 1
        self.daemon = True
        self.show_top_n = 10

    def _get_line(self, data) -> str:
        """ "Returns the top-n itens in data as a string"""
        return " | ".join(
            f"{k}: {v}" for k, v in data.items()[: self.show_top_n]
        )

    def run(self) -> None:
        while self.running:
            if self.stats_tracker is None:
                time.sleep(self.sleep_period)
                continue
            ready_queue_size = (
                len(self.stats_tracker.ready_queue)
                if self.stats_tracker.ready_queue is not None
                else 0
            )
            fmt = self._get_line
            self.logger.debug(
                f"Queue size: {ready_queue_size}\n"
                f"Call counts: {fmt(self.stats_tracker.call_counts)}\n"
                f"Max exec times: {fmt(self.stats_tracker.max_time)}\n"
                f"Total exec times: {fmt(self.stats_tracker.total_time)}\n"
                f"Avg exec times: {fmt(self.stats_tracker.avg_time)}"
            )
            time.sleep(self.sleep_period)

    def start(self) -> None:
        super().start()

    def stop(self):
        self.running = False



import asyncio
import platform
import re
import time
from unittest import mock

from asyncio_inspector.patcher import enable_inpection
from asyncio_inspector.reporter import LoggerReporter


async def do_nothing() -> int:
    """Async function that only sleeps(0) and return"""
    await asyncio.sleep(0)
    return 1


def test_logger_reporter():
    logger = mock.Mock()

    loop = asyncio.get_event_loop()
    reporter = LoggerReporter(logger=logger)
    reporter.sleep_period = 0.1
    reporter.start()
    with enable_inpection(loop, reporter=reporter):
        for _ in range(5):
            loop.call_soon(do_nothing)
        loop.call_soon(loop.stop)
        loop.run_forever()

    # Wait at least 2x the sleep_period
    time.sleep(0.2)
    reporter.stop()

    # And make sure the reporter logged at least once
    assert logger.debug.call_count >= 1
    last_log_call = logger.debug.call_args_list[-1]
    msg = last_log_call.args[0]
    assert "Call counts: do_nothing: 5 |" in msg

    assert re.search(r"Max exec times: do_nothing: \d+ ", msg, re.MULTILINE)
    assert re.search(r"Total exec times: do_nothing: \d+ ", msg, re.MULTILINE)
    assert re.search(
        r"Avg exec times: do_nothing: \d+(\.\d+)? ", msg, re.MULTILINE
    )

    if platform.system() == "Windows":
        assert "Queue size: 1\n" in msg
    else:
        assert "Queue size: 0\n" in msg



from typing import Optional

from sortedcollections import ValueSortedDict

from asyncio_inspector.events import ObservableDeque, ObservableHandle


def invert(x):
    return -x


class BaseStatsTracker:
    """"Basic stats tracker""" ""

    ready_queue: Optional[ObservableDeque]
    call_counts: ValueSortedDict
    total_time: ValueSortedDict
    avg_time: ValueSortedDict
    max_time: ValueSortedDict
    min_time: ValueSortedDict

    def __init__(self, ready_queue: ObservableDeque = None) -> None:
        self.ready_queue = ready_queue
        self.call_counts = ValueSortedDict(invert)
        self.total_time = ValueSortedDict(invert)
        self.avg_time = ValueSortedDict(invert)
        self.max_time = ValueSortedDict(invert)
        self.min_time = ValueSortedDict(invert)

    def track_call(
        self,
        handle: ObservableHandle,
        start_timestamp: int,
        end_timestamp: int,
    ) -> None:
        callback = handle.get_callback()
        if callback not in self.call_counts:
            self.call_counts[callback] = 0
            self.total_time[callback] = 0
            self.avg_time[callback] = 0
            self.max_time[callback] = 0
            self.min_time[callback] = 0

        self.call_counts[callback] += 1
        elapsed_time = end_timestamp - start_timestamp
        self.total_time[callback] += elapsed_time
        if elapsed_time > self.max_time[callback]:
            self.max_time[callback] = elapsed_time
        if elapsed_time < self.min_time[callback]:
            self.min_time[callback] = elapsed_time
        self.avg_time[callback] = (
            self.total_time[callback] / self.call_counts[callback]
        )



import time
from asyncio import Handle
from collections import deque
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from asyncio_inspector.stats import BaseStatsTracker


class ObservableHandle(Handle):
    """Subclass of handle that enables observability"""

    __slots__ = ("stats_tracker",)

    stats_tracker: "BaseStatsTracker"

    @classmethod
    def from_handle(cls, obj: Handle, stats_tracker) -> "ObservableHandle":
        """"Creates a new ObservableHandle from the given Handle object""" ""
        obs_handle = cls.__new__(cls)
        attrs_to_copy = [
            i for i in Handle.__slots__ if i != "__weakref__"  # type: ignore
        ]
        for attr in attrs_to_copy:
            value = getattr(obj, attr)
            setattr(obs_handle, attr, value)
        obs_handle.stats_tracker = stats_tracker
        return obs_handle

    def _run(self, *args, **kwargs):
        """Executes the original handle, tracking statistics"""
        start = time.time_ns()
        super(ObservableHandle, self)._run(*args, **kwargs)
        end = time.time_ns()
        self.stats_tracker.track_call(self, start, end)

    def get_callback(self) -> str:
        """"Returns the string representation of the callback""" ""
        callback = self._callback  # type: ignore
        # Check for TaskStepMethWrapper builtin type, which is used in
        # asyncio.run method to wrap the original coro.
        if hasattr(callback, "__self__") and hasattr(
            callback.__self__, "get_coro"
        ):
            callback = callback.__self__.get_coro().__qualname__
        elif hasattr(callback, "__qualname__"):
            callback = callback.__qualname__
        elif hasattr(callback, "__name__"):
            callback = callback.__name__
        else:
            # Last resource, just to make sure we track something
            callback = str(callback)
        return callback


class ObservableDeque(deque):
    """A deque of ObservableHandle"""

    __slots__ = ("stats_tracker",)

    def __init__(self, *args, **kwargs):
        super(ObservableDeque, self).__init__(*args, **kwargs)
        self.stats_tracker = None

    def popleft(self, *args, **kwargs):
        handle = super(ObservableDeque, self).popleft(*args, **kwargs)
        return ObservableHandle.from_handle(handle, self.stats_tracker)



import asyncio

from asyncio_inspector import enable_inpection, inspect, uninspect
from asyncio_inspector.events import ObservableDeque


async def do_nothing() -> int:
    """Async function that only sleeps(0) and return"""
    await asyncio.sleep(0)
    return 1


def test_patch_even_loop() -> None:
    """Makes sure we can patch an event loop and track call_soon calls"""
    loop = asyncio.get_event_loop()
    try:
        stats_tracker = inspect(loop)
        assert isinstance(loop._ready, ObservableDeque)
        loop.call_soon(do_nothing)
        loop.call_soon(loop.stop)
        loop.run_forever()
    finally:
        uninspect(loop)
        assert not isinstance(loop._ready, ObservableDeque)
    assert "do_nothing" in stats_tracker.call_counts
    assert stats_tracker.call_counts["do_nothing"] == 1


def test_patch_event_loop_context_manager() -> None:
    """ "Makes sure we can patch even loop with a context manager"""
    loop = asyncio.get_event_loop()
    with enable_inpection(loop) as stats_tracker:
        assert isinstance(loop._ready, ObservableDeque)
        loop.call_soon(do_nothing)
        loop.call_soon(loop.stop)
        loop.run_forever()

    assert "do_nothing" in stats_tracker.call_counts
    assert stats_tracker.call_counts["do_nothing"] == 1
    assert not isinstance(loop._ready, ObservableDeque)


def test_patch_event_loop_from_within() -> None:
    async def main():
        loop = asyncio.get_event_loop()
        with enable_inpection(loop) as stats_tracker:
            assert isinstance(loop._ready, ObservableDeque)
            await do_nothing()
        assert not isinstance(loop._ready, ObservableDeque)
        return stats_tracker

    stats_tracker = asyncio.run(main())
    method_name = "test_patch_event_loop_from_within.<locals>.main"
    assert method_name in stats_tracker.call_counts
    assert stats_tracker.call_counts[method_name] == 1



from asyncio import AbstractEventLoop
from collections import deque
from contextlib import contextmanager
from typing import Any, Iterator, Tuple

from asyncio_inspector.events import ObservableDeque
from asyncio_inspector.stats import BaseStatsTracker


def patch_event_loop_handler_creator(
    event_loop: AbstractEventLoop, stats_tracker: BaseStatsTracker
) -> None:
    obs_deque = ObservableDeque(event_loop._ready)  # type: ignore
    obs_deque.stats_tracker = stats_tracker
    stats_tracker.ready_queue = obs_deque
    event_loop._ready = obs_deque  # type: ignore


def unpatch_event_loop_handler_creator(event_loop: AbstractEventLoop) -> None:
    event_loop._ready = deque(event_loop._ready)  # type: ignore


@contextmanager
def enable_inpection(
    event_loop: AbstractEventLoop,
    *,
    stats_tracker=None,
    reporter=None,
) -> Iterator:
    """Patches the given event loop to enable inspection."""
    stats_tracker, reporter = _wire_reporter_and_stats_tracker(
        stats_tracker, reporter
    )
    patch_event_loop_handler_creator(event_loop, stats_tracker)
    yield stats_tracker
    unpatch_event_loop_handler_creator(event_loop)


def inspect(
    event_loop: AbstractEventLoop, *, stats_tracker=None, reporter=None
) -> BaseStatsTracker:
    stats_tracker, reporter = _wire_reporter_and_stats_tracker(
        stats_tracker, reporter
    )
    patch_event_loop_handler_creator(event_loop, stats_tracker)
    return stats_tracker


def uninspect(event_loop: AbstractEventLoop) -> None:
    unpatch_event_loop_handler_creator(event_loop)


def _wire_reporter_and_stats_tracker(
    stats_tracker, reporter
) -> Tuple[Any, Any]:
    if stats_tracker is None:
        stats_tracker = BaseStatsTracker()
    if reporter is not None:
        reporter.stats_tracker = stats_tracker
    return stats_tracker, reporter
